/**
 *  Copyright (c) 2020, Texas Instruments Incorporated
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:\
 *
 *  *   Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer.
 *  notice, this list of conditions and the following disclaimer in the
 *  documentation and/or other materials provided with the distribution.
 *  *   Neither the name of Texas Instruments Incorporated nor the names of
 *  its contributors may be used to endorse or promote products derived
 *  from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 *  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 *  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 *  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
import { AbstractDataCodec } from '../../ti-target-configuration/lib/AbstractCodec';
import { binaryOrBufferDataType, binaryDataType } from '../../ti-target-configuration/lib/CodecDataTypes';
import { streamingCodecDataType } from '../../ti-model-streaming/lib/StreamingDataModel';
import { IConfigurationAttributes } from '../../ti-target-configuration/lib/IConfigurationAttributes';
import { TiPromise, IDeferedPromise } from '../../ti-core-assets/lib/TiPromise';
import { BootLoader, IDataReader, IDataSender, ISerialPort, setLogger, IBinaryFileLoader } from './BootLoader';
import { TiFiles } from '../../ti-core-assets/lib/TiFiles';
import { UsbTransport } from '../../ti-transport-usb/lib/UsbTransport';
import { IUsbSerialPort } from '../../ti-service-usb/lib/UsbService';
import { TiConsole } from '../../ti-core-assets/lib/TiConsole';

/* eslint-disable @typescript-eslint/no-explicit-any */

const MODULE_NAME = 'boolloader';
setLogger('info', TiConsole.info.bind(TiConsole, MODULE_NAME));
setLogger('debug', TiConsole.debug.bind(TiConsole, MODULE_NAME));
setLogger('error', TiConsole.error.bind(TiConsole, MODULE_NAME));
TiConsole.setLevel(MODULE_NAME, 4);

// TODO: remove this when TiFiles stops supporting Blob.
function convertBlob2ArrayBuffer(data: Blob) {
    return new Promise<ArrayBuffer>((resolve) => {
        const reader = new FileReader();
        reader.onloadend = () => {
            resolve(reader.result as ArrayBuffer);
        };
        reader.readAsArrayBuffer(data);
    });
};

const binaryFileLoader = new (class implements IBinaryFileLoader {
    async loadBinaryFile(filepath: string) {
        return new Uint8Array(await convertBlob2ArrayBuffer(await TiFiles.readBinaryFile(filepath)));
    }
})();

export class BootLoaderCodec extends AbstractDataCodec<number[] | Buffer, number[], object, object> implements IDataReader, IDataSender, ISerialPort {
    private bootLoader: BootLoader;

    constructor(params: IConfigurationAttributes) {
        super(params.id || 'bootloader', binaryOrBufferDataType, binaryDataType, streamingCodecDataType, streamingCodecDataType);

        this.bootLoader = new BootLoader(this, this, 15, () => {
        });
    };

    encode() {
    };

    private partialPacket?: Uint8Array;

    private deferedReadPromise?: IDeferedPromise<number[]>;
    private requiredDataLength = 0;

    decode(rxData: number[] | Buffer): boolean | Error {

        let result: boolean | Error = false;
        // concatentate new data with saved partial data
        if (this.partialPacket) {
            this.partialPacket = new Uint8Array([...this.partialPacket, ...rxData]);
        } else {
            this.partialPacket = new Uint8Array(rxData);
        }

        if (this.partialPacket.length >= this.requiredDataLength && this.deferedReadPromise) {
            this.deferedReadPromise.resolve();
            result = true;
        }
        return result;
    };

    async readData(size: number, timeout: number) {
        if (!this.partialPacket || this.partialPacket.length < size) {
            this.requiredDataLength = size;
            this.deferedReadPromise = TiPromise.defer<number[]>();

            let promise = this.deferedReadPromise.promise;
            if (timeout > 0) {
                promise = TiPromise.timeout(promise, timeout * 1000, 'Timeout on reading data from serial port');
            };
            try {
                await promise;
            } finally {
                this.requiredDataLength = 0;
                this.deferedReadPromise = undefined;
            }
        }
        const result = this.partialPacket!.slice(0, size);
        this.partialPacket = this.partialPacket!.slice(size);

        return result;
    };

    sendData(data: Uint8Array) {
        if (this.targetEncoder) {
            this.targetEncoder.encode([...data]);
        }
    };

    private abortAllOperations = false;
    protected abort() {
        // make sure all queued operations, that are waiting to go, are aborted too
        this.abortAllOperations = true;
        this.sequencer = this.sequencer.finally(() => {
            this.abortAllOperations = false;
        });

        // reject the current operation
        if (this.deferedReadPromise) {
            this.deferedReadPromise.reject('This operation was aborted by the user.');
            this.deferedReadPromise = undefined;
        }
        this.flushInput();

        return this.sequencer;
    };

    private sequencer = Promise.resolve();   // sequencer for queuing operations to perform one after the other.
    private debouncePromises = new Map<string, Promise<any> | undefined>();
    private debounceAndSequenceOperation<T>(name: string, operation: (...params: any) => Promise<T>): () => Promise<T> {
        return (...params: any) => {
            let result = this.debouncePromises.get(name);
            if (!result) {
                result = this.sequencer.finally(() => {
                    if (this.abortAllOperations) {
                        throw new Error('This operation was aborted by the user.');
                    }
                }).then(() => {
                    return operation(...params);
                });
                this.debouncePromises.set(name, result);
                this.sequencer = result.finally(() => {
                    this.debouncePromises.set(name, undefined);
                });
            }
            return result;
        };
    };

    refresh: () => Promise<void> = this.debounceAndSequenceOperation<void>('refresh', async () => {
        const fl = await this.bootLoader.get_file_list();
        const info = await this.bootLoader.fs_usage_info();

        fl.sort();
        this.targetDecoder.decode({ userFiles: fl, info: info });
    });

    downloadFile: (filepath: string) => Promise<Uint8Array> = this.debounceAndSequenceOperation<Uint8Array>('download', async (filepath: string) => {
        return await this.bootLoader.dev_mode_download_fs_file(filepath);
    });

    uploadFile: (filepath: string, contents: string | Uint8Array) => Promise<void> = this.debounceAndSequenceOperation<void>('upload', async (filepath: string, contents: Uint8Array | string) => {
        await this.bootLoader.dev_mode_upload_fs_file(filepath, {}, contents);
    });

    setBreak(value: boolean): void {
        if ((this.targetEncoder as UsbTransport).selectedPort) {
            const port = (this.targetEncoder as UsbTransport).selectedPort;
            if ((port as IUsbSerialPort).setSignals) {
                (port as IUsbSerialPort).setSignals({ brk: value });
            }
        }
    };

    flushInput() {
        this.requiredDataLength = 0;
        this.partialPacket = undefined;
    };

    async connect() {
        await this.abort();
        const data = await this.bootLoader.connect_device_api(this, binaryFileLoader);
        if (this.targetDecoder) {
            this.targetDecoder.decode(data);
        }
    };

    disconnect() {
        return this.bootLoader.disconnect();
    };
};
